# Program to convert a non-negative integer x into its binary
# represention

# Reads an integer n and extracts its binary representation

# Pre: integer n must be non-negative and < 2^nbits 
# Post: bit i of n will be placed into bits[i]

# idiom to initialize PC with START addr
_:START;

START: ;

# bits posn    0   1 ... (nbits-1)
# bits value 2^0 2^1 ... 2^(nbits-1)

# Maximum number of bits
nbits = 8;

# Begin by computing a table of the power of 2 associated with each
# bit position.

# Pointer p_pow2 to bit 0, incremented on each bit
SUB, pow2_start_addr, _0, p_pow2;

# Current power of 2
SUB, _1, _0, two_to_i;

# loop i = 0 .. n-1 times, using i as the index
SUB, _0, _0, i;
loop_1_begin: BGE, i, n, loop_1_end;

    # indirect store:  [two_to_i] -> [ [p_pow2] ]
    # build instruction to store pow2 into that address
    SUB, p_pow2, _0, @+7;
    SUB, two_to_i, _0, 0;

    # double the power of 2
    SUB, _0, two_to_i, tmp;
    SUB, two_to_i, tmp, two_to_i;

    # increment i and address of next pow2 table entry
    SUB, i, _n1, i;
    SUB, p_pow2, _n1, p_pow2;

    # back to top of loop
    BGE, _0, _0, loop_1_begin;
loop_1_end:

# Now read in a value and determine its bin rep
SUB, IN, _0, value;
SUB, value, _0, remain;

# end address is 1 beyond actual table, and we want to start
# at last entries and work down

SUB, pow2_end_addr, _1, p_pow2;
SUB, bits_end_addr, _1, p_bits;

# Work in decreasing powers of 2
# loop i = 0 .. n-1 times, using i as the index

SUB, _0, _0, i;

loop_2_begin: BGE, i, n, loop_2_end;

    # indirect load:  [ [p_pow2] ] -> [ two_to_i ]
    # build instruction to load 2^(n-1-i) into two_to_i
    SUB, p_pow2, _0, @+5;
    SUB, 0, _0, two_to_i;

    # remain < pow => remain does not contain that power so bit is 0
    # otw bit is 1

    SUB, _0, _0, bit;
    BGE, remain, two_to_i, bigger;
    BGE, _0, _0, smaller;
    bigger: 
        SUB, _1, _0, bit;
        # remove that power from remainder
        SUB, remain, two_to_i, remain;
    smaller:

    # debug
    # SUB, i, _0, OUT;
    # SUB, a, _0, OUT;
    # SUB, b, _0, OUT;
    # SUB, two_to_i, _0, OUT;
    # SUB, bit, _0, OUT;
    # SUB, remain, _0, OUT;
    # SUB, _n1, _0, OUT;

    # build instruction to store the bit into the bits table.
    SUB, p_bits, _0, @+7;
    SUB, bit, _0, 0;

    # increment i and decrement address of current entry in power
    # table and bit value
    SUB, i, _n1, i;
    SUB, p_pow2, _1, p_pow2;
    SUB, p_bits, _1, p_bits;

    # back to top of loop
    BGE, _0, _0, loop_2_begin;
loop_2_end:

# Halt by setting PC to 0
SUB, _0, _0, PC;

# Some constants.  Note that if you put them in this order
# then you can index off _0, for example, this stores 1 - (-1) 
# into tmp.
#   SUB, _0+1, _0-1, tmp;

_n2: -2;
_n1: -1;
_0: 0;
_1: 1;
_2: 2;

# Input value to be converted
value: 0;

# Current power 2^n
two_to_i: 1;

# Remaining bits of value to be processed, should be 0 when
# finished.
remain: 0;

# Current bit value
bit: 0;

# Variables

# Number of bits expected
n: nbits; 

# Index
i: 0;

# Pointer to pow2 entry
p_pow2: 0;

# Pointer to bits entry
p_bits: 0;

# Volatile tmp value
tmp: 0;

# the idiom to reserve space for a table of size k is
#
# table_start_addr: table - addr of element 0
# table_end_addr: table_end - 1 past element k
#
# table: ;
# @ = @ + size;
# table_end: ;

# power table is [pow2_start_addr ... pow2_end_addr] inclusive
pow2_start_addr: pow2;
pow2_end_addr: pow2_end;

# bits table is [bits_start_addr ... bits_end_addr - 1] 
bits_start_addr: bits;
bits_end_addr: bits_end;

# powers of 2 table starts here, pow2[i] = 2^i 0 <= i < nbits
pow2: ;
@=@+nbits;
pow2_end:;

# bits of binary representation start here, bits[i] = value of bit i,
# in little endian ordering. 
#   bit[i] is value of bit representing 2^i

bits: ;
@=@+nbits;
bits_end: ;
